<?php


if (!defined("ABSPATH")) {
    exit;
}

/**
 * 
 * Simple logger class.
 *
 * Simple class with static methods & properties to:
 *   - Keep track of log entries.
 *   - Keep track of timings with time() and timeEnd() methods, Javascript style.
 *   - Optionally write log entries in real-time to file or to screen.
 *   - Optionally dump the log to file in one go at any time.
 * 
 * Log entries can be added with any of the following methods:
 *  - info( $message, $title = '' )      // an informational message intended for the user
 *  - debug( $message, $title = '' )     // a diagnostic message intended for the developer
 *  - warning( $message, $title = '' )   // a warning that something might go wrong
 *  - error( $message, $title = '' )     // explain why the program is going to crash
 * The $title argument is optional; if present, it will be
 * prepended to the message: "$title => $message'.
 * 
 * For example, the following code
 *  > Logger::info( "program started" );
 *  > Logger::debug( "variable x is false" );
 *  > Logger::warning( "variable not set, something bad might happen" );
 *  > Logger::error( "file not found, exiting" );
 * will print to STDOUT the following lines:
 *  $ 2021-07-21T11:11:03+02:00 [INFO] : program started
 *  $ 2021-07-21T11:11:03+02:00 [DEBUG] : variable x is false
 *  $ 2021-07-21T11:11:03+02:00 [WARNING] : variable not set, something bad might happen
 *  $ 2021-07-21T11:11:03+02:00 [ERROR] : file not found, exiting
 *
 * To write to file, prepend the following line:
 *  > Logger::$write_log = true;
 *
 * To customize the log file path:
 *  > Logger::$log_dir = 'mylogdir';
 *  > Logger::$log_file_name = 'logname';
 *  > Logger::$log_file_extension = 'txt';
 *
 * To overwrite the log file at every run of the script:
 *  > Logger::$log_file_append = false;
 *
 * To prevent printing to STDOUT:
 * > Logger::$print_log = false;
 *
 * Note: the function uses the $logger_ready property to understand whether to load
 * the init() function. We do this to make the class work straight away, without the
 * need to instantiate it. This however can create race conditions if you are executing
 * parallel code. 
 *
 * TODO: Remove the static methods and make this an actual class that can be initialized
 * multiple times.
 * 
 */
if ( ! class_exists( 'BindCommerce_Logger' ) ) {

    /**
     * 
     * The main bindCommerce Logger class
     *
     * @since 1.0.1
     * 
     */ 
    class BindCommerce_Logger extends BindCommerce {
                
        /**
         * 
         * Incremental log, where each entry is an array with the following elements:
         *
         *  - timestamp => timestamp in seconds as returned by time()
         *  - level => severity of the bug; one between debug, warning, error, critical
         *  - name => name of the log entry, optional
         *  - message => actual log message
         * 
         * @since 1.0.1
         * 
         */
        protected static $log = [];

        /**
         * Whether to print log entries to screen as they are added.
         * 
         * @since 1.0.1
         * 
         */
        public static $print_log = true;

        /**
         * Whether to write log entries to file as they are added.
         * 
         * @since 1.0.1
         * 
         */
        public static $write_log = false;

        /**
         * Directory where the log will be dumped, without final slash; default
         * is this file's directory
         * 
         * @since 1.0.1
         * 
         */
        public static $log_dir = BINDCOMMERCE_LOGS_PATH;
        
        /**
         * File name for the log saved in the log dir
         * 
         * @since 1.0.1
         * 
         */
        public static $log_file_name = "bindcommerce_log";

        /**
         * File extension for the logs saved in the log dir
         * 
         * @since 1.0.1
         * 
         */
        public static $log_file_extension = "log";
        
        /**
         * Whether to append to the log file (true) or to overwrite it (false)
         * 
         * @since 1.0.1
         * 
         */
        public static $log_file_append = true;

        /**
         * Absolute path of the log file, built at run time
         * 
         * @since 1.0.1
         * 
         */
        private static $log_file_path = '';

        /**
         * Where should we write/print the output to? Built at run time
         * 
         * @since 1.0.1
         * 
         */
        private static $output_streams = [];

        /**
         * Whether the init() function has already been called
         * 
         * @since 1.0.1
         * 
         */
        private static $logger_ready = false;
        
        /**
         * Associative array used as a buffer to keep track of timed logs
         * 
         * @since 1.0.1
         * 
         */
        private static $time_tracking = [];


        /**
         * Add a log entry with an informational message for the user.
         * 
         * @since 1.0.1
         * 
         */
        public static function info( $message, $name = '', $send_email = false ) {

            // Send notify to email if enabled
            if ( BindCommerce_Setting::is_debug_notify_email_enabled() 
                && !empty(BindCommerce_Setting::get_debug_notify_email_list()) 
                && $send_email) 
            {

                $emails = BindCommerce_Setting::get_debug_notify_email_list();
                
                if (strpos($emails, ',') !== false) {

                    $emails = explode(",", $emails); 

                } else {

                    $arr = [];
                    $arr[] = $emails;
                    $emails = $arr;

                }

                foreach ($emails as $email) {

                    $to = $email;
                    $subject = __("WP - BindCommerce INFO Log", BINDCOMMERCE_PLUGIN_TEXT_DOMAIN);
                    $body = $message;
                    $headers = array('Content-Type: text/html; charset=UTF-8');
                    
                    wp_mail( $to, $subject, $body, $headers );

                } 

            }

            return self::add( $message, $name, 'info' );

        }


        /**
         * Add a log entry with a diagnostic message for the developer.
         * 
         * @since 1.0.1
         * 
         */
        public static function debug( $message, $name = '', $send_email = false ) {
            
            // Send notify to email if enabled
            if ( BindCommerce_Setting::is_debug_notify_email_enabled() 
                && !empty(BindCommerce_Setting::get_debug_notify_email_list()) 
                && $send_email) 
            {

                $emails = BindCommerce_Setting::get_debug_notify_email_list();
                
                if (strpos($emails, ',') !== false) {

                    $emails = explode(",", $emails); 

                } else {

                    $arr = [];
                    $arr[] = $emails;
                    $emails = $arr;

                }

                foreach ($emails as $email) {

                    $to = $email;
                    $subject = __("WP - BindCommerce DEBUG Log", BINDCOMMERCE_PLUGIN_TEXT_DOMAIN);
                    $body = $message;
                    $headers = array('Content-Type: text/html; charset=UTF-8');
                    
                    wp_mail( $to, $subject, $body, $headers );

                } 

            }

            return self::add( $message, $name, 'debug' );
        }


        /**
         * Add a log entry with a warning message.
         * 
         * @since 1.0.1
         * 
         */
        public static function warning( $message, $name = '', $send_email = false ) {
            
            // Send notify to email if enabled
            if ( BindCommerce_Setting::is_debug_notify_email_enabled() 
                && !empty(BindCommerce_Setting::get_debug_notify_email_list()) 
                && $send_email ) 
            {

                $emails = BindCommerce_Setting::get_debug_notify_email_list();
                
                if (strpos($emails, ',') !== false) {

                    $emails = explode(",", $emails); 

                } else {

                    $arr = [];
                    $arr[] = $emails;
                    $emails = $arr;

                }

                foreach ($emails as $email) {

                    $to = $email;
                    $subject = __("WP - BindCommerce WARNING Log", BINDCOMMERCE_PLUGIN_TEXT_DOMAIN);
                    $body = $message;
                    $headers = array('Content-Type: text/html; charset=UTF-8');
                    
                    wp_mail( $to, $subject, $body, $headers );

                } 

            }

            return self::add( $message, $name, 'warning' );
        }


        /**
         * Add a log entry with an error - usually followed by
         * script termination.
         */
        public static function error( $message, $name = '', $send_email = false ) {
            
            // Send notify to email if enabled
            if ( BindCommerce_Setting::is_debug_notify_email_enabled() 
                && !empty(BindCommerce_Setting::get_debug_notify_email_list()) 
                && $send_email ) 
            {

                $emails = BindCommerce_Setting::get_debug_notify_email_list();
                
                if (strpos($emails, ',') !== false) {

                    $emails = explode(",", $emails); 

                } else {

                    $arr = [];
                    $arr[] = $emails;
                    $emails = $arr;

                }

                foreach ($emails as $email) {

                    $to = $email;
                    $subject = __("WP - BindCommerce ERROR Log", BINDCOMMERCE_PLUGIN_TEXT_DOMAIN);
                    $body = $message;
                    $headers = array('Content-Type: text/html; charset=UTF-8');
                    
                    wp_mail( $to, $subject, $body, $headers );

                } 

            }

            return self::add( $message, $name, 'error' );

        }


        /**
         * Start counting time, using $name as identifier.
         *
         * Returns the start time or false if a time tracker with the same name
         * exists
         * 
         * @since 1.0.1
         * 
         */
        public static function time( string $name ) {

            if ( ! isset( self::$time_tracking[ $name ] ) ) {
                self::$time_tracking[ $name ] = microtime( true );
                return self::$time_tracking[ $name ];
            }
            else {
                return false;
            }
        }


        /**
         * Stop counting time, and create a log entry reporting the elapsed amount of
         * time.
         *
         * Returns the total time elapsed for the given time-tracker, or false if the
         * time tracker is not found.
         * 
         * @since 1.0.1
         * 
         */
        public static function timeEnd( string $name ) {

            if ( isset( self::$time_tracking[ $name ] ) ) {
                $start = self::$time_tracking[ $name ];
                $end = microtime( true );
                $elapsed_time = number_format( ( $end - $start), 2 );
                unset( self::$time_tracking[ $name] );
                self::add( "$elapsed_time seconds", "'$name' took", "timing" );
                return $elapsed_time;
            }
            else {
                return false;
            }
        }


        /**
         * Add an entry to the log.
         *
         * This function does not update the pretty log. 
         * 
         * @since 1.0.1
         * 
         */
        private static function add( $message, $name = '', $level = 'debug' ) {

            /* Create the log entry */
            $log_entry = [
                'timestamp' => time(),
                'name' => $name,
                'message' => $message,
                'level' => $level,
            ];

            /* Add the log entry to the incremental log */
            self::$log[] = $log_entry;

            /* Initialize the logger if it hasn't been done already */
            if ( ! self::$logger_ready ) {
                self::init();
            }

            /* Write the log to output, if requested */
            if ( self::$logger_ready && count( self::$output_streams ) > 0 ) {
                $output_line = self::format_log_entry( $log_entry ) . PHP_EOL;
                foreach ( self::$output_streams as $key => $stream ) {
                    fputs( $stream, $output_line );
                }
            }

            return $log_entry;
        }


        /**
         * Take one log entry and return a one-line human readable string
         * 
         * @since 1.0.1
         * 
         */
        public static function format_log_entry( array $log_entry ) : string {

            $log_line = "";

            if ( ! empty( $log_entry ) ) {

                /* Make sure the log entry is stringified */
                $log_entry = array_map( function( $v ) { return print_r( $v, true ); }, $log_entry );
            
                /* Build a line of the pretty log */
                $log_line .= date( 'c', $log_entry['timestamp'] ) . " ";
                $log_line .= "[" . strtoupper( $log_entry['level'] ) . "] : ";
                if ( ! empty( $log_entry['name'] ) ) {
                    $log_line .= $log_entry['name'] . " => ";
                }
                $log_line .= $log_entry['message'];
            
            }
            
            return $log_line;
        }
        
        
        /**
         * Determine whether an where the log needs to be written; executed only
         * once.
         *
         * @return {array} - An associative array with the output streams. The 
         * keys are 'output' for STDOUT and the filename for file streams.
         * 
         * @since 1.0.1
         * 
         */
        public static function init() {

            if ( ! self::$logger_ready ) {

                /* Print to screen */
                if ( true === self::$print_log ) {
                    self::$output_streams[ 'stdout' ] = STDOUT;
                }

                /* Create directory if not exist */
                if ( !file_exists( self::$log_dir ) ) {
                    mkdir(self::$log_dir, 0777, true);
                }

                /* Build log file path */
                if ( file_exists( self::$log_dir ) ) {
                    self::$log_file_path = implode( DIRECTORY_SEPARATOR, [ self::$log_dir, self::$log_file_name ] );
                    if ( ! empty( self::$log_file_extension ) ) {
                        self::$log_file_path .= "." . self::$log_file_extension;
                    }
                }

                /* Print to log file */
                //if ( true === self::$write_log ) {
                if (BindCommerce_Setting::is_debug_mode()) {
                    if ( file_exists( self::$log_dir ) ) {
                        $mode = self::$log_file_append ? "a" : "w";
                        self::$output_streams[ self::$log_file_path ] = fopen ( self::$log_file_path, $mode );  
                    }
                }
            }

            /* Now that we have assigned the output stream, this function does not need
            to be called anymore */
            self::$logger_ready = true;

        }


        /**
         * Dump the whole log to the given file.
         *
         * Useful if you don't know before-hand the name of the log file. Otherwise,
         * you should use the real-time logging option, that is, the $write_log or
         * $print_log options.
         *
         * The method format_log_entry() is used to format the log.
         *
         * @param {string} $file_path - Absolute path of the output file. If empty,
         * will use the class property $log_file_path.
         * 
         * @since 1.0.1
         * 
         */
        public static function dump_to_file( $file_path='' ) {

            if ( ! $file_path ) {
                $file_path = self::$log_file_path;
            }
            
            if ( file_exists( dirname( $file_path ) ) ) {

                $mode = self::$log_file_append ? "a" : "w";
                $output_file = fopen( $file_path, $mode );

                foreach ( self::$log as $log_entry ) {
                    $log_line = self::format_log_entry( $log_entry );
                    fwrite( $output_file, $log_line . PHP_EOL );
                }
                
                fclose( $output_file );
            }
        }
        
        
        /**
         * Dump the whole log to string, and return it.
         *
         * The method format_log_entry() is used to format the log.
         * 
         * @since 1.0.1
         * 
         */
        public static function dump_to_string() {
        
            $output = '';
            
            foreach ( self::$log as $log_entry ) {
                $log_line = self::format_log_entry( $log_entry );
                $output .= $log_line . PHP_EOL;
            }
            
            return $output;
        }


        /**
         * 
         * Get logs and returns them 
         * 
         * @since 1.0.1
         * 
         */
        public static function get_logs( $limit=10, $file_path='', $type = "string" ) {

            if ( ! $file_path ) {
                $file_path = implode( DIRECTORY_SEPARATOR, [ self::$log_dir, self::$log_file_name ] ) . "." . self::$log_file_extension;
            }
            
            if ( file_exists( dirname( $file_path ) ) ) {

                $lines = [];
                $logs = [];
                $log_string = "";
                $handle = fopen($file_path, "r");

                if ($handle) {

                    while (($line = fgets($handle)) !== false) {

                        // process the line read.
                        $lines[] = $line;

                    }

                    if (!empty($lines)) {

                        $lines = array_reverse($lines);

                        if ($limit > count($lines) ) {
                            $limit = count($lines);
                        }
                        
                        for ($i = 0; $i < $limit; $i++) {
    
                            $logs[] = $lines[$i];
    
                        }

                        $logs = array_reverse($logs);
                        $lines = [];

                        foreach ($logs as $log) {
    
                            $lines[] = $log;
                            $log_string .= $log . "\r\n";

                        }

                        $logs = $lines;
    
                    }

                    fclose($handle);

                    if ($type == "string") {
                              
                        if (!empty($log_string)) {
                            $log_string = str_replace("\r\n\r\n", "<br />", $log_string);
                            $log_string = str_replace("\r\n", "<br />", $log_string);
                        }
                        return $log_string;

                    }
                    if ($type == "array") {
                        return (array) $logs;
                    }
                    if ($type == "object") {
                        return (object) $logs;
                    }
                } 

            }

            return false;
            
        }


        /**
         * 
         * Delete logs 
         * 
         * @since 1.0.1
         * 
         */
        public static function delete_logs( $file_path='') {

            if ( ! $file_path ) {
                $file_path = implode( DIRECTORY_SEPARATOR, [ self::$log_dir, self::$log_file_name ] ) . "." . self::$log_file_extension;
            }
            
            if ( file_exists( dirname( $file_path ) ) ) {

                $contents = file_get_contents($file_path);
                $contents = "";
                file_put_contents($file_path, $contents);
                return true;

            }

            return false;
            
        }

    } 

}