<?php

if ( ! defined( 'ABSPATH' ) ) { 
    exit; 
}


if ( ! class_exists( 'BindCommerce_Product' ) ) {

    /**
     * 
     * BindCommerce_Product class 
     * 
     * @since 1.0.0
     * 
     */
    class BindCommerce_Product extends BindCommerce{

        public $_wpdb;          // Wordpress global queries passed in the construct
        public $_wc_product;    // WC Product Object passed in the construct
        public $_product_id;    // Product ID passed in the construct


        /**
         * 
         * Main constructor
         * 
         * @since 1.0.0
         * 
         */
        public function __construct( $product_id ){

            global $wpdb;
            $this->_wpdb = $wpdb;

            $this->_wc_product = wc_get_product( $product_id );
            $this->_product_id = $product_id;
            
        }


        /**
         * 
         * Returns the product type according to WooCommerce
         * 
         * @since 1.0.0
         * 
         */
        public function get_product_type() {

            if( $this->_wc_product->is_type( 'simple' ) ){ 
                return 'simple';            
            }

            if( $this->_wc_product->is_type( 'variable' ) ){ 
                return 'variable';            
            }

            if( $this->_wc_product->is_type( 'variation' ) ){ 
                return 'variation';            
            }

        }


        /**
         * 
         * Returns product sync status specified in the bindcommerce tab for each product
         * 
         * @since 1.0.0
         * 
         */
        public function get_product_sync_status() {

            $sync_status = get_post_meta( $this->_product_id, 'bindcommerce_product_sync', true );

            if ( $sync_status == "yes" || $sync_status == "1" ) {
                return false;
            }

            if (get_post_status($this->_product_id) == 'draft' && !BindCommerce_Setting::get_export_draft_products_status()) {
                return false;
            }

            return true;

        }
        

        /**
         * 
         * Returns current currency according to WooCommerce
         * 
         * @since 1.0.0
         * 
         */
        public function get_wc_currency() {

           return get_woocommerce_currency();

        }

        /**
         * 
         * Returns all attributes linked to the product according to WooCommerce
         * 
         * @since 1.0.0
         * 
         */
        public function get_product_attributes() {
            
            $arr = [];
            $attributes = $this->_wc_product->get_attributes();            

            foreach ( $attributes as $attribute_key => $attribute ) {

                $attribute_label        = wc_attribute_label( $attribute->get_name() );

                $manufacturer_label     = BindCommerce_DB::get_result("

                    SELECT attribute_label 
                    FROM " . DB_NAME . "." . BINDCOMMERCE_DB_WP_PREFIX . BINDCOMMERCE_WC_ATTRIBUTE_TAXONOMIES . " " . "
                    WHERE attribute_name = '". BindCommerce_Setting::get_product_manufacturer_mapping() . "';

                "); 

                $supplier_label         = BindCommerce_DB::get_result("

                    SELECT attribute_label 
                    FROM " . DB_NAME . "." . BINDCOMMERCE_DB_WP_PREFIX . BINDCOMMERCE_WC_ATTRIBUTE_TAXONOMIES . " " . "
                    WHERE attribute_name = '". BindCommerce_Setting::get_product_supplier_mapping() . "';

                "); 

                if ( $attribute_label != $manufacturer_label && $attribute_label != $supplier_label ) {

                    if (!$attribute->get_variation()) {

                        $arr["Attribute"][] = [    
        
                            "Name"       => wc_attribute_label( $attribute->get_name() ),
                            "Value"      => $this->_wc_product->get_attribute( $attribute->get_name() ),    
        
                        ];
                        
                    }

                }
                
            }    

            if ($this->_wc_product->get_tax_class()) {                

                $arr["Attribute"][] = [    

                    "Name"       => "VAT Rule",
                    "Value"      => $this->_wc_product->get_tax_class(),    

                ];     

            }

            return $arr;

        }


        /**
         * 
         * Returns all attributes linked to the variation product according to WooCommerce
         * 
         * @since 1.0.0
         * 
         */
        public function get_variation_attributes() {
            
            $arr = [];
            $attributes     = array_keys( $this->_wc_product->get_variation_attributes( false ) );                      

            foreach ($attributes as $attribute) {

                $attribute_label        = wc_attribute_label( $attribute );
                $manufacturer_label     = BindCommerce_DB::get_result("

                    SELECT attribute_label 
                    FROM " . DB_NAME . "." . BINDCOMMERCE_DB_WP_PREFIX . BINDCOMMERCE_WC_ATTRIBUTE_TAXONOMIES . " " . "
                    WHERE attribute_name = '". BindCommerce_Setting::get_product_manufacturer_mapping() . "';

                "); 

                $supplier_label         = BindCommerce_DB::get_result("

                    SELECT attribute_label 
                    FROM " . DB_NAME . "." . BINDCOMMERCE_DB_WP_PREFIX . BINDCOMMERCE_WC_ATTRIBUTE_TAXONOMIES . " " . "
                    WHERE attribute_name = '". BindCommerce_Setting::get_product_supplier_mapping() . "';

                "); 

                if ( $attribute_label != $manufacturer_label && $attribute_label != $supplier_label ) {

                    $arr["Attribute"][] = [    

                        "Name"       => wc_attribute_label( $attribute ),
                        "Value"      => $this->_wc_product->get_attribute($attribute)   

                    ];

                }

            }    
            return $arr;

        }


        /**
         * 
         * Returns main image linked to the product according to WooCommerce
         * 
         * @since 1.0.0
         * 
         */
        public function get_product_main_image() {

            $product_main_image = $this->_wc_product->get_image_id();
            return wp_get_attachment_url( $product_main_image );            

        }

        /**
         * 
         * Returns all images linked to the product according to WooCommerce
         * 
         * @since 1.0.0
         * 
         */
        public function get_product_images() {

            $arr = [];
            $product_main_image         = $this->_wc_product->get_image_id();
            $product_gallery_image_ids  = $this->_wc_product->get_gallery_image_ids();

            $arr["Picture"][]["URL"] = wp_get_attachment_url( $product_main_image );

            foreach( $product_gallery_image_ids as $product_image_id ) {

                $arr["Picture"][]["URL"] = wp_get_attachment_url( $product_image_id );

            }    
            
            return $arr;           

        }
        

        /**
         * 
         * Returns all categories ID linked to the product according to WooCommerce
         * 
         * @since 1.0.0
         * 
         */
        public function get_product_categories_id() {
            return $this->_wc_product->get_category_ids();
        }
       

        /**
         * 
         * Returns all alternatives according to WPML and WooCommerce
         * 
         * @since 1.0.0
         * 
         */
        public function get_product_alternatives() {

            $arr = array();

            if ( BindCommerce_Addons::is_wpml_active() ) { 

                global $sitepress;
                $translated_ids = array();

                if(!isset($sitepress)) {

                    return array();

                }
                
                $trid = $sitepress->get_element_trid($this->_product_id, 'post_product');
                $translations = $sitepress->get_element_translations($trid, 'product');
                
                foreach( $translations as $lang=>$translation){

                    $translated_ids[] = $translation->element_id;

                }

                foreach( $translated_ids as $translated_product_id){

                    if ($this->_product_id == $translated_product_id || empty($translated_product_id) ) {
                        continue;
                    }

                    $wpml_product               = wc_get_product($translated_product_id); 
                    $language_code              = bindcommerce_get_product_language_code($translated_product_id, true);                    
                    $product_title              = $wpml_product->get_title();
                    $product_html_description   = $wpml_product->get_description();
                    $product_short_description  = $wpml_product->get_short_description();
                    $product_notes              = $wpml_product->get_purchase_note();

                    $arr["Alternative"][] = [

                        "Language"          => $language_code,
                        "Title"             => $product_title,
                        "DescriptionHtml"   => $product_html_description,
                        "ShortDescription"  => $product_short_description,
                        "Notes"             => $product_notes,
    
                    ];   
                    

                }

                return $arr;         
                       
            
            } 

            return array();

        }


        /**
         * 
         * Returns all categories and subcategories name linked to the product according to WooCommerce
         * 
         * @since 1.0.0
         * 
         */
        public function get_product_categories() {

            $cat_arr = [];
            $cont = 0;
            $cats = bindcommerce_wc_get_parent_categories();

            if ( BindCommerce_Addons::is_wpml_active() ) { 

                $language_code = apply_filters( 'wpml_post_language_details', NULL, $this->_product_id );
                $language_code = $language_code["language_code"];            
            
                foreach ($cats as $cat) {
            
                    $wpml_category_id = apply_filters( 'wpml_object_id', $cat->cat_ID, 'product_cat', false, $language_code);

                    if($cat->parent==0) {
            
                        if( bindcommerce_has_term_relationship( $this->_product_id, $wpml_category_id ) ) {

                            $cat_arr["Category"][$cont] = bindcommerce_wpml_get_category_name( $wpml_category_id );

                        }
                        
                        $arg=[[],[]];
                        bindcommerce_wc_get_subcategories($cat,$arg);

                        if ($arg[0]) {

                            $l=0;
                            foreach ($arg[0] as $scat) {

                                $i=$arg[1][$l];
                                $wpml_category_id = apply_filters( 'wpml_object_id', $scat->cat_ID, 'product_cat', false, $language_code);
            
                                if( bindcommerce_has_term_relationship( $this->_product_id, $wpml_category_id ) ) {
                                    
                                    if (!empty($cat_arr["Category"][$cont])) {   

                                        $cat_arr["Category"][$cont] .= "|" . bindcommerce_wpml_get_category_name($wpml_category_id);

                                    } else {
                                                                  
                                        $cat_arr["Category"][$cont] .= bindcommerce_wpml_get_category_name($wpml_category_id);

                                    }

                                }
            
                                $l++;
                            }

                        }
                        
                    }
                    $cont++;
                }      

                $cat_arr = [
                    "Category" => array_unique($cat_arr["Category"]) 
                ];                      
            
            } else {

                foreach ($cats as $cat) {
    
                    if($cat->parent==0) {
    
                        if( has_term( $cat->name, 'product_cat', $this->_product_id ) ) {
                            $cat_arr["Category"][$cont] = $cat->name;
                        }
                        
                        $arg=[[],[]];
                        bindcommerce_wc_get_subcategories($cat,$arg);
                        if ($arg[0]) {
                            $l=0;
                            foreach ($arg[0] as $scat) {
                                $i=$arg[1][$l];
    
                                if( has_term( $scat->name, 'product_cat', $this->_product_id ) ) {

                                    if (!empty($cat_arr["Category"][$cont])) {   

                                        $cat_arr["Category"][$cont] .= "|" . $scat->name;

                                    } else {

                                        $cat_arr["Category"][$cont] .= $scat->name;

                                    }
                                    
                                }
    
                                $l++;
                            }
                        }
                        
                    }
                    $cont++;
                }

            }

            return $cat_arr;

        }


        /**
         * 
         * Returns product regular price
         * 
         * @since 1.0.0
         * 
         */
        public function get_product_regular_price() {

            if ( $this->get_product_type() == 'simple' ) {

                $regular_price = get_post_meta( $this->_product_id, BINDCOMMERCE_PRODUCT_META_SIMPLE_REGULAR_PRICE, true );

            }

            if ( $this->get_product_type() == 'variation' ) {

                $regular_price = get_post_meta( $this->_product_id, BINDCOMMERCE_PRODUCT_META_VARIATION_REGULAR_PRICE, true );

            }

            if ( !empty($regular_price) ) {
                return $regular_price;
            }   
            return $this->_wc_product->get_regular_price();    

                  

        }


        /**
         * 
         * Returns product sale status
         * 
         * @since 1.0.0
         * 
         */
        public function get_product_sale_status( $is_numeric=false ) {

            if ( !$is_numeric ) {

                if ( $this->_wc_product->is_on_sale() ) {
                    return true;
                }   
                return false;  

            }     
            
            if ( $is_numeric ) {

                if ( $this->_wc_product->is_on_sale() ) {
                    
                    if ( get_option( 'woocommerce_calc_taxes' ) === "yes" ) {
                        
                        if ( get_option( 'woocommerce_prices_include_tax' ) === "no" ) {

                            return -1; 
        
                        } else {    
                        
                            return 1;
        
                        }

                    }
                    return 1;

                }   
                return 0;  

            }    

            return false;

        }


        /**
         * 
         * Returns product sale price
         * 
         * @since 1.0.0
         * 
         */
        public function get_product_sale_price() {

            if ( $this->get_product_type() == 'simple' ) {

                $sale_price = get_post_meta( $this->_product_id, BINDCOMMERCE_PRODUCT_META_SIMPLE_SALE_PRICE, true );

            }

            if ( $this->get_product_type() == 'variation' ) {

                $sale_price = get_post_meta( $this->_product_id, BINDCOMMERCE_PRODUCT_META_VARIATION_SALE_PRICE, true );

            }

            if ( !empty($sale_price) ) {
                return $sale_price;
            }   
            return $this->_wc_product->get_sale_price();                     

        }


        /**
         * 
         * Returns product net price
         * 
         * @since 1.0.0
         * 
         */
        public function get_product_net_price() {

            $product_net_price          = $this->get_product_regular_price(); 
            $product_supplement_amount  = $this->get_product_supplement_amount();
                
            if ($product_supplement_amount) {

                if (strpos($product_supplement_amount, '%') !== false) {

                    $product_supplement_amount         = str_replace("%", "", $product_supplement_amount);
                    $product_supplement_amount         = ($product_supplement_amount + 100) / 100;
                    return $product_net_price * $product_supplement_amount;

                } else {

                    return $product_net_price + $product_supplement_amount;

                }

            } 
            return $product_net_price;                  

        }


        /**
         * 
         * Returns product gross price
         * 
         * @since 1.0.0
         * 
         */
        public function get_product_gross_price() {

            $product_gross_price        = $this->get_product_net_price(); 
            $product_tax_amount         = $this->get_product_tax_amount();

            if ($product_tax_amount) {

                $product_tax_amount     = str_replace("%", "", $product_tax_amount);                

                if ( get_option( 'woocommerce_prices_include_tax' ) === "no" ) {

                    return ($product_gross_price * $product_tax_amount) / 100;   

                } else {

                    return $product_gross_price - ( ($product_gross_price * 100) / (100 + $product_tax_amount) );   

                }

            }
            
            return "";    

        }


        /**
         * 
         * Returns product tax amount percentage
         * 
         * @since 1.0.0
         * 
         */
        public function get_product_tax_amount_percentage() {

            if ( $this->get_product_type() == 'simple' ) {

                $tax_amount = get_post_meta( $this->_product_id, BINDCOMMERCE_PRODUCT_META_SIMPLE_TAX_AMOUNT, true );

            }  
            
            if ( $this->get_product_type() == 'variation' ) {

                $tax_amount = get_post_meta( $this->_product_id, BINDCOMMERCE_PRODUCT_META_VARIATION_TAX_AMOUNT, true );
                
            } 

            if ( !empty($tax_amount) ) {
                return $tax_amount;
            }  

            if ( BindCommerce_Setting::get_tax_amount_status() ) {

                $tax_amount_setting = BindCommerce_Setting::get_tax_amount();
                if ( !empty($tax_amount_setting) ) {
                    return $tax_amount_setting;
                }

            }
            return false;

        }


        /**
         * 
         * Returns product tax amount
         * 
         * @since 1.0.0
         * 
         */
        public function get_product_tax_amount($decimal = false) {
            
            if ( BindCommerce_Setting::get_tax_amount_status() && !empty($this->get_product_tax_amount_percentage()) ) {

                if ($decimal) {

                    return $this->get_product_tax_amount_percentage() / 100;

                }
                return $this->get_product_tax_amount_percentage();

            } else {

                if ( get_option( 'woocommerce_calc_taxes' ) === "yes" && $this->_wc_product->get_tax_class() ) {
                
                    $tax    = new WC_Tax();
                    $taxes  = $tax->get_rates($this->_wc_product->get_tax_class());

                    if ($taxes) {

                        if (!is_array($taxes)) {
                            $taxes = (array) $taxes;
                        }

                        $rates  = array_shift($taxes);
                        //Take only the item rate and round it. 
                        $wc_tax_amount = round(array_shift($rates));
    
                        if ($decimal) {
    
                            return $wc_tax_amount / 100;
        
                        }
    
                        return $wc_tax_amount;   
                        
                    }

                }
                return false; 

            }
            return false;            

        }


        /**
         * 
         * Returns product supplement amount
         * 
         * @since 1.0.0
         * 
         */
        public function get_product_supplement_amount() {

            if ( $this->get_product_type() == 'simple' ) {

                $supplement_amount = get_post_meta( $this->_product_id, BINDCOMMERCE_PRODUCT_META_SIMPLE_SUPPLEMENT_AMOUNT, true );

            }   

            if ( $this->get_product_type() == 'variation' ) {

                $supplement_amount = get_post_meta( $this->_product_id, BINDCOMMERCE_PRODUCT_META_VARIATION_SUPPLEMENT_AMOUNT, true );

            }   

            if ( !empty($supplement_amount) ) {
                return $supplement_amount;
            }   

            if ( BindCommerce_Setting::get_supplement_amount_status() ) {

                $supplement_amount_setting = BindCommerce_Setting::get_supplement_amount();
                if ( !empty($supplement_amount_setting) ) {
                    return $supplement_amount_setting;
                }

            }

            return false;


        }


        /**
         * 
         * Returns product prices according to bindCommerce API
         * 
         * @since 1.0.0
         * 
         */
        public function get_product_prices() {

            $arr = [];
            $product_currency           = $this->get_wc_currency();    
            $product_net_price          = $this->get_product_net_price();   
            $product_sale_status        = $this->get_product_sale_status(true);
            $product_sale_price         = $this->get_product_sale_price();

            $product_tax_amount         = 0;
            if ( $this->get_product_tax_amount() ) {
                $product_tax_amount     = $this->get_product_tax_amount(true);
            }

            $product_gross_price        = 0;
            if ( $this->get_product_gross_price() ) {
                $product_gross_price    = $this->get_product_gross_price();
            }

            if ( BindCommerce_Setting::get_tax_amount_status() ) {

                if ( $product_gross_price == 0 ) {
                    $product_gross_price = $product_net_price;
                }

                // bindCommerce TAX Option activated and set
                $arr["Price"][0] = [

                    "ListCode"      => "Woo",
                    "ListName"      => "WooCommerce",
                    "Currency"      => $product_currency,
                    "Vat"           => $product_tax_amount,
                    "NetPrice"      => $product_net_price,
                    "GrossPrice"    => $product_gross_price,
                    "OverridePrice" => $product_sale_price,
                    "Override"      => $product_sale_status,

                ];

            } else {
                        
                $rates = BindCommerce_DB::get_results("

                    SELECT *
                    FROM " . DB_NAME . "." .  BINDCOMMERCE_DB_WP_PREFIX . BINDCOMMERCE_WC_DB_WOOCOMMERCE_TAX_RATES . " " . "
                    WHERE tax_rate_class='" . $this->_wc_product->get_tax_class() ."';
                    
                ");   

                if ( $rates && get_option( 'woocommerce_calc_taxes' ) === "yes" ) {

                    foreach ($rates as $rate) {

                        $product_net_price          = floatval($this->get_product_net_price());   
                        $product_gross_price        = floatval($this->get_product_net_price());  
                        $product_tax_amount         = floatval($rate->tax_rate);

                        if ($product_tax_amount) {

                            if ( get_option( 'woocommerce_prices_include_tax' ) === "no" ) {

                                $product_gross_price = $product_net_price + ( ($product_net_price * $product_tax_amount) / 100 );   

                            } else {

                                $product_gross_price = $product_net_price;   
                                $product_net_price   = ( ($product_gross_price * 100) / (100 + $product_tax_amount) );   

                            }

                            $product_tax_amount  = $product_tax_amount / 100;     

                        }

                        $arr["Price"][] = [

                            "ListCode"      => "Woo" . "-" . $rate->tax_rate_country, // TODO: Check if $rate->tax_rate_country is empty
                            "ListName"      => "WooCommerce",
                            "Currency"      => $product_currency,
                            "Vat"           => $product_tax_amount,
                            "NetPrice"      => $product_net_price,
                            "GrossPrice"    => $product_gross_price,
                            "OverridePrice" => $product_sale_price,
                            "Override"      => $product_sale_status,
    
                        ];

                    }

                } else {

                    // No rates available

                    if ( $product_gross_price == 0 ) {
                        $product_gross_price = $product_net_price;
                    }

                    $arr["Price"][0] = [

                        "ListCode"      => "Woo",
                        "ListName"      => "WooCommerce",
                        "Currency"      => $product_currency,
                        "Vat"           => $product_tax_amount,
                        "NetPrice"      => $product_net_price,
                        "GrossPrice"    => $product_gross_price,
                        "OverridePrice" => $product_sale_price,
                        "Override"      => $product_sale_status,

                    ];

                }

            }

            return $arr;

        }

        


        /**
         * 
         * Returns product quantity
         * 
         * @since 1.0.0
         * 
         */
        public function get_product_quantity() {

            if ( $this->get_product_type() == 'simple' ) {

                $stock_quantity = get_post_meta( $this->_product_id, BINDCOMMERCE_PRODUCT_META_SIMPLE_QUANTITY, true );

            }            

            if ( $this->get_product_type() == 'variation' ) {

                $stock_quantity = get_post_meta( $this->_product_id, BINDCOMMERCE_PRODUCT_META_VARIATION_QUANTITY, true );

            }

            if ( !empty($stock_quantity) ) {
                return $stock_quantity;
            }   

            $product_stock_quantity = $this->_wc_product->get_stock_quantity();
            
            if (!$product_stock_quantity && $product_stock_quantity !== 0) {
                $product_stock_quantity = 999;
            }

            return $product_stock_quantity;

        }


        /**
         * 
         * Returns product GTIN 
         * 
         * @since 1.0.0
         * 
         */
        public function get_product_gtin() {            

            if ( BindCommerce_Setting::get_gtin_fields_status() ) {
                
                if ( $this->get_product_type() == 'simple' ) {

                    return get_post_meta( $this->_product_id, BINDCOMMERCE_PRODUCT_META_SIMPLE_GTIN, true );

                }
                
                if ( $this->get_product_type() == 'variation' ) {

                    return get_post_meta( $this->_product_id, BINDCOMMERCE_PRODUCT_META_VARIATION_GTIN, true );

                }

            }
            return false;       
            
        }


        /**
         * 
         * Returns product Manufacturer 
         * 
         * @since 1.0.0
         * 
         */
        public function get_product_manufacturer() {            

            if ( BindCommerce_Setting::get_product_manufacturer_mapping() ) {
                
                if ( BindCommerce_Addons::is_wpml_active() ) {

                    global $sitepress;
                                    
                    $attribute_name     = "pa_" . BindCommerce_Setting::get_product_manufacturer_mapping();
                    $term_name          = $this->_wc_product->get_attribute( BindCommerce_Setting::get_product_manufacturer_mapping() );
                    $manufacturer_id    = bindcommerce_get_attribute_id_by_name( $term_name );

                    $language_code      = apply_filters( 'wpml_post_language_details', NULL, $this->_product_id );
                    $language_code      = $language_code["language_code"];      

                    $term_trid          = $sitepress->get_element_trid( $manufacturer_id, 'tax_' . $attribute_name );
                    $term_translations  = $sitepress->get_element_translations( $term_trid, 'tax_' . $attribute_name , false, false, true ); 

                    $term_translation  = $term_translations[$language_code];                   

                    $manufacturer = $term_translation->name;
                    
                } else {

                    $manufacturer = $this->_wc_product->get_attribute( BindCommerce_Setting::get_product_manufacturer_mapping() );

                }           
                

                return $manufacturer;

            }
            return false;       
            
        }


        /**
         * 
         * Returns product Supplier 
         * 
         * @since 1.0.0
         * 
         */
        public function get_product_supplier() {            

            if ( BindCommerce_Setting::get_product_supplier_mapping() ) {

                if ( BindCommerce_Addons::is_wpml_active() ) {

                    global $sitepress;
                                    
                    $attribute_name     = "pa_" . BindCommerce_Setting::get_product_supplier_mapping();
                    $term_name          = $this->_wc_product->get_attribute( BindCommerce_Setting::get_product_supplier_mapping() );
                    $supplier_id        = bindcommerce_get_attribute_id_by_name( $term_name );

                    $language_code      = apply_filters( 'wpml_post_language_details', NULL, $this->_product_id );
                    $language_code      = $language_code["language_code"];      

                    $term_trid          = $sitepress->get_element_trid( $supplier_id, 'tax_' . $attribute_name );
                    $term_translations  = $sitepress->get_element_translations( $term_trid, 'tax_' . $attribute_name , false, false, true ); 

                    $term_translation   = $term_translations[$language_code];                   

                    $supplier = $term_translation->name;
                    
                } else {

                    $supplier = $this->_wc_product->get_attribute( BindCommerce_Setting::get_product_supplier_mapping() );

                }                           

                return $supplier;

            }
            return false;       
            
        }


        
        /**
         * 
         * Get the products trace ready to send
         * 
         * @since 1.0.0
         * 
         */
        public static function get_products_trace( $params ){  
        
            $ids_arr    = [];
            $bc_arr     = [];

            $product_create_sync_codes = [];
            $product_delete_sync_codes = [];
    
            $products_id        = (array) $params["products_id"];
            $products_limit     = $params["limit"];
            $products_order_by  = $params["order_by"];
            $mode               = $params["mode"];
            $debug              = $params["debug"];
            $type               = $params["type"];

            if (empty($mode)) {
                $mode = "full";
            }

            if (empty($type)) {
                $type = "incremental";
            }
    
            foreach ($products_id as $product_id) {

                if (!is_numeric($product_id)) {
                    return false;
                }
                
            }

            if ( $products_order_by == "asc" || $products_order_by == "ASC" ) {

                asort($products_id);

            } 

            if ( $products_limit == "-1" || empty($products_limit) ) {

                $products_limit = count($products_id);

            } 

            $limit_count = 0;            

            foreach ($products_id as $product_id) {

                if ($limit_count == $products_limit) {
                    break;
                }               
    
                $product = new BindCommerce_Product( $product_id );
                $product_variants = [];
    
                if ($product->get_product_sync_status() && (get_post_status($product_id) == 'publish' || get_post_status($product_id) == 'draft') ) {            
                    
                    $bc_product_prefix = "";  
                    $product_code      = "";
                    
                    if ( BindCommerce_Setting::get_product_prefix_status() ) {
    
                        $bc_product_prefix = BindCommerce_Setting::get_product_prefix();    
                    
                    }
    
                    if ( BindCommerce_Setting::is_product_mode_id() ) {
    
                        $product_code = $product_id;
    
                    }
    
                    if ( BindCommerce_Setting::is_product_mode_sku() ) {
                        
                        $product_code = $product->_wc_product->get_sku();
    
                    }
    
                    if ($product_code == "") {
    
                        $product_code = $product_id;
    
                    }
                        
                    $ids_arr[] = $product_id;

                    $product_create_sync_codes[] = [ 
                        $product_id => $product_code 
                    ];
                    
                    $language_code = bindcommerce_get_product_language_code($product_id, true);
                        
                    $product_title              = $product->_wc_product->get_title();
                    $product_html_description   = $product->_wc_product->get_description();
                    $product_short_description  = $product->_wc_product->get_short_description();
                    $product_notes              = $product->_wc_product->get_purchase_note();
    
                    $product_quantity           = $product->get_product_quantity();
                    $product_alternatives       = $product->get_product_alternatives(); 
                    $product_categories         = $product->get_product_categories(); 
                    $product_images             = $product->get_product_images(); 
                    $product_attributes         = $product->get_product_attributes();  
                    $product_prices             = $product->get_product_prices();   
                    $product_gtin               = $product->get_product_gtin(); 

                    $product_manufacturer       = $product->get_product_manufacturer(); 
                    $product_supplier           = $product->get_product_supplier(); 
    
                    $product_weight             = $product->_wc_product->get_weight();  
                    $product_length             = $product->_wc_product->get_length();      
                    $product_width              = $product->_wc_product->get_width();      
                    $product_height             = $product->_wc_product->get_height();   
    
                    if ($product->get_product_type() == "variable") {
    
                        foreach ( $product->_wc_product->get_children( false ) as $child_id ) {
                            
                            $variation = new BindCommerce_Product( $child_id );
    
                            if ( !$variation->_wc_product || !$variation->_wc_product->exists() ) {
    
                                continue;
    
                            }
    
                            $variation_code = "";
                            
                            if ( BindCommerce_Setting::is_product_mode_id() ) {
    
                                $variation_code           = $child_id;
            
                            }
            
                            if ( BindCommerce_Setting::is_product_mode_sku() ) {
                                
                                $variation_code           = $variation->_wc_product->get_sku();
            
                            }                        
    
                            if ($variation_code == "") {
    
                                $variation_code           = $child_id;
        
                            }
    
                            $variation_title              = $variation->_wc_product->get_name(); // output "Product name - Variation Attribute name"
                            $variation_html_description   = $variation->_wc_product->get_description();
                            $variation_notes              = $variation->_wc_product->get_purchase_note();
        
                            $variation_quantity           = $variation->get_product_quantity();
                            $variation_images             = $variation->get_product_images();  
                            $variation_attributes         = $variation->get_variation_attributes();  
                            $variation_prices             = $variation->get_product_prices();   
                            $variation_gtin               = $variation->get_product_gtin(); 
    
                            $variation_weight             = $variation->_wc_product->get_weight();  
                            $variation_length             = $variation->_wc_product->get_length();      
                            $variation_width              = $variation->_wc_product->get_width();      
                            $variation_height             = $variation->_wc_product->get_height();   

                            if ($mode == "full") {
    
                                $product_variants["Variant"][] = [
                                        "Language"              => $language_code,
                                        "Code"                  => $bc_product_prefix . $variation_code,
                                        "Barcode"               => $variation_gtin,
                                        //"BarcodeKind"           => "EAN",
                                        "Title"                 => $variation_title,
                                        "DescriptionHtml"       => $variation_html_description,
                                        "Notes"                 => $variation_notes,
                                        "Prices"                => $variation_prices,
                                        "Dimensions"            => [
                                            "Weight"                    => $variation_weight,
                                            "WeightUom"                 => "kg", // TODO: Check in db
                                            "WeightVolume"              => "", // TODO: Check in db
                                            "LwhUom"                    => "cm", // TODO: Check in db
                                            "Length"                    => $variation_length,
                                            "Width"                     => $variation_width,
                                            "Height"                    => $variation_height
                                        ],
                                        "Qty"                   => $variation_quantity,
                                        "Attributes"            => $variation_attributes,
                                        "Pictures"              => $variation_images                            
                                ];

                            } 

                            if ($mode == "quantity") {
    
                                $product_variants["Variant"][] = [
                                        "Language"              => $language_code,
                                        "Code"                  => $bc_product_prefix . $variation_code,
                                        "Qty"                   => $variation_quantity                           
                                ];

                            } 

                            if ($mode == "price") {
    
                                $product_variants["Variant"][] = [
                                        "Language"              => $language_code,
                                        "Code"                  => $bc_product_prefix . $variation_code,
                                        "Prices"                => $variation_prices             
                                ];

                            } 
    
    
                        }
    
                    }
    
                    if ($mode == "full") {

                        $bc_arr["Products"]["Product"][] = [
                            "Language"              => $language_code,
                            "Code"                  => $bc_product_prefix . $product_code,
                            "Barcode"               => $product_gtin,
                            //"BarcodeKind"           => "EAN",
                            "Title"                 => $product_title,
                            "DescriptionHtml"       => $product_html_description,
                            "ShortDescription"      => $product_short_description,
                            "Notes"                 => $product_notes,
                            "Categories"            => $product_categories,
                            "Alternatives"          => $product_alternatives,
                            "Prices"                => $product_prices,
                            "Manufacturer"          => $product_manufacturer,
                            "MPN"                   => "",
                            "Supplier"              => $product_supplier,
                            "Dimensions"            => [
                                "Weight"                    => $product_weight,
                                "WeightUom"                 => "kg", // TODO: Check in db
                                "WeightVolume"              => "", // TODO: Check in db
                                "LwhUom"                    => "cm", // TODO: Check in db
                                "Length"                    => $product_length,
                                "Width"                     => $product_width,
                                "Height"                    => $product_height
                            ],
                            "Qty"                   => $product_quantity,
                            "Pictures"              => $product_images,
                            "Attributes"            => $product_attributes,
                            "Warehouses"            => [],
                            "Variants"              => $product_variants,
                            "Derivatives"           => []                         
                        ];

                    }

                    if ($mode == "quantity") {

                        $bc_arr["Products"]["Product"][] = [
                            "Language"              => $language_code,
                            "Code"                  => $bc_product_prefix . $product_code,
                            "Qty"                   => $product_quantity,
                            "Warehouses"            => [],
                            "Variants"              => $product_variants,
                            "Derivatives"           => []                        
                        ];
                        
                    }

                    if ($mode == "price") {

                        $bc_arr["Products"]["Product"][] = [
                            "Language"              => $language_code,
                            "Code"                  => $bc_product_prefix . $product_code,
                            "Prices"                => $product_prices,
                            "Warehouses"            => [],
                            "Variants"              => $product_variants,
                            "Derivatives"           => []                        
                        ];

                    }
    
                } else {

                    // Delete products on bindCommerce

                    $bc_product_prefix = "";
                    $product_code      = "";
                    
                    if ( BindCommerce_Setting::get_product_prefix_status() ) {
    
                        $bc_product_prefix = BindCommerce_Setting::get_product_prefix();
                    
                    }

                    if ( BindCommerce_Setting::is_product_mode_sku() ) {
                        
                        $product_code           = $product->_wc_product->get_sku();
    
                    }
    
                    if ($product_code == "") {
    
                        $product_code           = $product_id;
    
                    }
                    if ($product_code == "") {
    
                        $product_code           = $product_id;
    
                    }
                        
                    $ids_arr[] = $product_id;
                    $bc_arr["Delete"]["sku"][] = $bc_product_prefix . $product_code;

                    $product_delete_sync_codes[] = [
                        $product_id => $bc_product_prefix . $product_code
                    ];
                
                }

                $limit_count++;                            
    
            }
                       
            $products_arr = "";

            if ($params["format"] == "application/json") {
    
                $products_arr = json_encode($bc_arr);
    
            }

            if ($params["format"] == "text/xml") {
    
                $products_arr = bindcommerce_array_to_xml($bc_arr, "bindCommerceProducts", $type);
    
            }

            if ($debug == true) {
    
                return $products_arr;
    
            }

            return [
                "postfields"                    => $products_arr,
                "ids"                           => $ids_arr,
                "product_create_sync_codes"     => $product_create_sync_codes,
                "product_delete_sync_codes"     => $product_delete_sync_codes,
            ];
        
            

        }


    }

}

?>